/* This program dump data from an Unidos device*/
/* using the RS232 interface. The results are printed in two coulm    */
/* ASCII in the form ((float) TIME [sec] ; (float) VOLTAGE [V] )      */ 
/* 04.06.2003 Niels Bassler, <bassler@phys.au.dk> */
/* */
/* Compiling: gcc get_serial.c -o get_serial */ 
/* last change 12.09.2009 /NB */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include "pbar.h"

#define DEBUG 0  /* debugging */
//#define DEBUG 1  /* debugging */


/* LeCroy things */
#define COMMAND_STRING_SIZE 100

/* RS232 Stuff */
//#define TTY "/dev/ttyS0"
#define TTY "/dev/ttyUSB0"
#define MAXBUF (256)
#define BUFFER_SIZE 0x100
#define RS232_SPEED B9600
#define RS232_TIMEOUT 10   // wait max 10 seconds

/* other parameters */
#define TRUE 1
#define FALSE 0

char *prog_name;

char *nb_fgets(char *s, int n, FILE *iop, char br);
int write_tty(char *command, char *answer, int len);
void SIGhandler(int i);
int open_tty(void);
int tty;

int main(int argc, char *argv[]) {

  char resp[BUFFER_SIZE];
  int resp_len;
  
  float result=0;
  float time = 0;

  int fd;
  int num;
  char s[BUFFER_SIZE];

  if (DEBUG)
    printf("open port\n");
  open_tty();
  
  if (DEBUG)
    printf("open port done\n");
  // test
  result = -1;
  
  resp_len = write_tty("G",resp,1);
  //  printf("resplen %i\n", resp_len);

  if (resp_len == 0) {
    result = -1;
    printf("%s: timeout.",argv[0]);
  }
  else if (resp_len == -1) {
    printf("%s: command too large.",argv[0]);
  }
  else {
    if (DEBUG)
      printf(">%s< -- ", resp);
  }


  sleep(UNIDOS_EXP_TIME); 
  resp_len = write_tty("V",resp,1);
  //printf("resplen %i\n", resp_len);

  
  if (resp_len == 0) {
    result = -1;
    printf("%s: timeout.",argv[0]);
  }
  else if (resp_len == -1) {
    printf("%s: command too large.",argv[0]);
  }
  else {
      if (DEBUG)
	printf(">%s< -- ", resp);
      sscanf(resp,"%fs %e %s", &time, &result, s);
      if (DEBUG)
	printf("t: %f val: %e\n", time, result);
  }

  close(tty);

  if (DEBUG)
    printf("closed tty\n");
  

  /* send data to fifo */
  sprintf(s,"%i %.6e ",DEV_UNIDOS, result);
  if (DEBUG)
    printf("open fifo\n");
  fd = open(FIFO_PBAR, O_WRONLY);
  if (fd == -1) {
    fprintf(stderr,"GET_SERIAL ERROR\n");
    fprintf(stderr,"Could not open FIFO_PBAR. Does it exist?");
    fprintf(stderr,"*mkfifo FIFO_PBAR\n");
    
    exit(-1);
  }
  if ((num = write(fd, s, strlen(s))) == -1)
    perror("get_serial");
  close(fd);
  if (DEBUG)
    printf("close fifo\n");

  return(1);
}





char *nb_fgets(char *s, int n, FILE *iop, char br) {
  /* modified fgets, which stops at breakpoint br */
  /* fgets would be br = '\n' */

  register int c = 0;
  register char *cs;
  
  cs = s;
  while (--n > 0 && (c = getc(iop)) != EOF)
    if ((*cs++ = c) == br)
      break;
    *cs = '\0';
  return (c == EOF && cs == s) ? NULL : s;
}
	

void SIGhandler(int i) {
        long int tty_stat;

        tty_stat=TIOCM_RTS+TIOCM_DTR;
        ioctl(tty,TIOCMBIC,&tty_stat);
        close(tty);
        fflush(stdout);
        exit(1);
}

int write_tty(char *command, char answer[], int len) {
  int i;
  char c;
  char temp_string[BUFFER_SIZE];
  
  // build string 


  if (len > (BUFFER_SIZE-2)) {
    printf("getserial: RS232 command too long.");
    return(-1);
  }

  for(i = 0; i < len; i++) 
    temp_string[i] = command[i];
  temp_string[i] = 0x0d; //cr
  temp_string[i+1] = 0x0a; //cr
  write(tty, temp_string, len+2);
  
  i = 0;

  while ((read(tty,&c,1) == 0) && (i < (RS232_TIMEOUT*10.0))) {
    // wait for response 100 msec
    usleep(100000);
    i++;    
  }

  /* if no timeout */
  if  (i < (RS232_TIMEOUT*10.0)) {    
    i = 0;
    answer[i++] = c;
    // and read the rest
    while( (read(tty,&c,1) > 0) && ( i< (BUFFER_SIZE-1))) 
      answer[i++] = c;
    answer[i] = '\0';
  }
  else
    i = 0;
  return(i); // length of string without null byte
}

int open_tty(void){
  /* returns tty */
  struct termios tios;
  int i;
  long int tty_stat;


  
  /*setup signal handling routine*/
  signal(SIGHUP,SIGhandler);
  signal(SIGINT,SIGhandler);
  signal(SIGQUIT,SIGhandler);
  signal(SIGTERM,SIGhandler);
  signal(SIGTSTP,SIGhandler);
  
  /*initialise tty port*/

  if (DEBUG)
    printf("    RS232 port settings: 8N1 9600 baud (hardcoded)\n");

  tios.c_iflag = IGNBRK + IGNPAR;
  tios.c_oflag = 0;
  tios.c_cflag = CREAD + CLOCAL + RS232_SPEED + CS8 + HUPCL ;
  tios.c_lflag = 0;
  for (i=0; i<NCCS; i++) {
    tios.c_cc[i]='\0';
  }
  tios.c_cc[VMIN]=0;
  
  if ((tty=open(TTY,O_RDWR))<0) {
    fprintf(stderr,"Cannot open %s\n",TTY);
    exit(1);
  }

  if (tcsetattr(tty, TCSANOW, &tios)) {
    fprintf(stderr,"Failed to initialise %s\n",TTY);
    exit(1);
  }
  tty_stat=TIOCM_RTS+TIOCM_DTR; /*See: modem(7) */
  ioctl(tty,TIOCMBIS,&tty_stat);
  return(1);
}

