/*  DESCRIPTION : 
    This little program must be forked from the parent program.
    It polls the parallel-port to see if there has been a trigger.
    If a trigger has been received, it will emit SIGUSR1 to the 
    parent process.
    After receiving a trigger signal, it will reset the flip-flop 
    in the trigger box.

    REQUIRES: 
    Parapin package. Custom build parallelport circuit.

    COMPILING: 
    gcc -Wall -O2 parport_trigger_daemon_v2.c -lparapin -o parport_trigger_daemon_v2
 
    NOTE: 
    This program must be run with root permissions.    
    change the ownership to root : 
    "chown root parport_trigger_daemon_v2"
    set the SUID i.e. : 
    "chmod u+s parport_trigger_daemon_v2"

    chown root parport_trigger_daemon_v2
    chmod u+s parport_trigger_daemon_v2
    chown root parport_trigger_daemon_v2_debug
    chmod u+s parport_trigger_daemon_v2_debug



    AUTHOR: 
    Niels Bassler, <n.bassler@dkfz.de>
    
    DATE: 
    05.10.2007

    CHANGELOG:
    05.10.2007
    - proper installation of libparapin 
    - kill if parent dies.

*/
#include <stdio.h>
#include <signal.h> 
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <parapin.h>

#define POLL_TIME           1000000   /* time between each poll in s */

#define TRIGGER_PULSE_WIDTH 200000  /* wait at least so long before 
					reseting and resume polling.
					This value must at least be as long
					as the trigger pulse. 
					No harm is done exaggerating here,
					so set this very large.
				     */


#define PIN_RESET       14 //3
#define PIN_DATA        16 //4
#define PIN_SET_DATA    17 //5
#define PIN_TRIGGER     11
//#define DEBUG           1        /* 1 = on, 
//				    0 = off */

int main(void)
{ 
#ifndef DEBUG
  int i;
#endif
  
  if (pin_init_user(LPT1) < 0)
    exit(0);
  /* initialisation and reset */ 
  pin_output_mode(LP_DATA_PINS | LP_SWITCHABLE_PINS);
  usleep(1000);
  clear_pin(LP_PIN[PIN_SET_DATA] | LP_PIN[PIN_DATA]);
  set_pin(LP_PIN[PIN_SET_DATA]);
  set_pin(LP_PIN[PIN_DATA]);
  /* Reset flip flop */
  clear_pin(LP_PIN[PIN_RESET]);
  usleep(1000000);
  set_pin(LP_PIN[PIN_RESET]);

  /* now the trigger pin should be low */
  // set all hig
/*   while(1) { */
/*     printf("on"); */
/*     sleep(4);fflush(stdout); */
/*     set_pin(LP_PIN[PIN_RESET]); */
/*     set_pin(LP_PIN[PIN_TRIGGER]); */
/*     set_pin(LP_PIN[PIN_SET_DATA]); */
/*     set_pin(LP_PIN[PIN_DATA]); */
/*     sleep(4); */
/*     printf("off");fflush(stdout); */
/*     clear_pin(LP_PIN[PIN_RESET]); */
/*     clear_pin(LP_PIN[PIN_TRIGGER]); */
/*     clear_pin(LP_PIN[PIN_SET_DATA]); */
/*     clear_pin(LP_PIN[PIN_DATA]); */
/*   } */

/*   exit(1); */
  

#ifdef DEBUG
    if(pin_is_set(LP_PIN[PIN_TRIGGER]))
      printf("pin 11 is set\n");
    else
      printf("pin 11 is low\n");
    
#endif
  /* (unless the Trigger box still is turned off) */

  /*
  if (DEBUG)
    printf(" parent process: %i", (int) getppid());
  */

    while(1) {
      usleep(POLL_TIME);

    /*
      if(DEBUG)
      printf("Pin 11: %s\n", pin_is_set(LP_PIN[PIN_TRIGGER]) ? "HIGH":"LOW");
    */
    if (pin_is_set(LP_PIN[PIN_TRIGGER])) {
#ifdef DEBUG
	printf("   Trigger daemon: got a valid trigger. \n");
#endif
      usleep(TRIGGER_PULSE_WIDTH); /* wait for trigger pulse to end for sure */
      clear_pin(LP_PIN[PIN_RESET]);
      usleep(1000);
      set_pin(LP_PIN[PIN_RESET]);

#ifndef DEBUG
      i = getppid();
      if (i == 1)
	exit(0);
      else
	kill( getppid(), SIGUSR1);
#endif
    }    
  }
  return(0);
}
