/*  DESCRIPTION : 
    This little program must be forked from the parent program.
    It polls the parallel-port to see if there has been a trigger.
    If a trigger has been received, it will emit SIGUSR1 to the 
    parent process.
    After receiving a trigger signal, it will reset the flip-flop 
    in the trigger box.

    REQUIRES: 
    Parapin package. Custom build parallelport circuit.

    COMPILING: 
    gcc -Wall -O2 parport_trigger_daemon_v2.c -lparapin -o parport_trigger_daemon_v2
 
    NOTE: 
    This program must be run with root permissions.    
    change the ownership to root : 
    "chown root parport_trigger_daemon_v2"
    set the SUID i.e. : 
    "chmod u+s parport_trigger_daemon_v2"

    chown root parport_trigger_daemon_v2
    chmod u+s parport_trigger_daemon_v2
    chown root parport_trigger_daemon_v2_debug
    chmod u+s parport_trigger_daemon_v2_debug
    chown root parport_trigger_daemon_v2_heikki
    chmod u+s parport_trigger_daemon_v2_heikki


    AUTHOR: 
    Niels Bassler, <bassler@phys.au.de>
    
    DATE: 
    05.10.2007

    CHANGELOG:
    05.10.2007
    - proper installation of libparapin 
    - kill if parent dies.

*/
#include <stdio.h>
#include <signal.h> 
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <parapin.h>
#include <fcntl.h>
#include <sys/stat.h> // S_IFIFO
#include "pbar.h"

//#define POLL_TIME           1000000   /* time between each poll in s */
// reduce to 0.2 seconds
#define POLL_TIME           200000   /* time between each poll in s */

#define TRIGGER_PULSE_WIDTH 200000  /* wait at least so long before 
					reseting and resume polling.
					This value must at least be as long
					as the trigger pulse. 
					No harm is done exaggerating here,
					so set this very large.
				     */

#define PIN_HEIKKI      1  // this pin for heikkis trigger.
#define PIN_RESET       14 //3
#define PIN_DATA        16 //4
#define PIN_SET_DATA    17 //5
#define PIN_TRIGGER     11
//#define DEBUG           1        /* 1 = on, 
//				    0 = off */
  
int main(void)
{ 
#ifndef DEBUG
  int i;
#endif
  int fd_fifo;
  int num;

  char readbuf[READBUF_SIZE];

  if (pin_init_user(LPT1) < 0)
    exit(0);
  /* initialisation and reset */ 
  pin_output_mode(LP_DATA_PINS | LP_SWITCHABLE_PINS);
  usleep(1000);
  clear_pin(LP_PIN[PIN_SET_DATA] | LP_PIN[PIN_DATA]);
  set_pin(LP_PIN[PIN_SET_DATA]);
  set_pin(LP_PIN[PIN_DATA]);
  /* Reset flip flop */
  clear_pin(LP_PIN[PIN_RESET]);
  usleep(1000000);
  set_pin(LP_PIN[PIN_RESET]);

  clear_pin(LP_PIN[PIN_HEIKKI]);  // set heikkis pin low


  /* now the trigger pin should be low */
  // set all hig
/*   while(1) { */
/*     printf("on"); */
/*     sleep(4);fflush(stdout); */
/*     set_pin(LP_PIN[PIN_RESET]); */
/*     set_pin(LP_PIN[PIN_TRIGGER]); */
/*     set_pin(LP_PIN[PIN_SET_DATA]); */
/*     set_pin(LP_PIN[PIN_DATA]); */
/*     sleep(4); */
/*     printf("off");fflush(stdout); */
/*     clear_pin(LP_PIN[PIN_RESET]); */
/*     clear_pin(LP_PIN[PIN_TRIGGER]); */
/*     clear_pin(LP_PIN[PIN_SET_DATA]); */
/*     clear_pin(LP_PIN[PIN_DATA]); */
/*   } */

/*   exit(1); */
  

#ifdef DEBUG
  if(pin_is_set(LP_PIN[PIN_TRIGGER]))
    printf("pin 11 is set\n");
  else
    printf("pin 11 is low\n");
  
#endif
  /* (unless the Trigger box still is turned off) */
  
  /*
    if (DEBUG)
    printf(" parent process: %i", (int) getppid());
  */
  
  

  /////// create named pipe for heikki
  
#ifdef DEBUG
  printf("create named pipe %s\n",FIFO_HEIKKI);
  fflush(stdout);
#endif

  umask(0);
  if(mknod(FIFO_HEIKKI, S_IFIFO|0666, 0) != 0) {
    unlink(FIFO_HEIKKI);
    if(mknod(FIFO_HEIKKI, S_IFIFO|0666, 0) != 0) {
      printf("Error: could not create fifo.\n"); 
      exit(-1);
    }
  }
  /* open pipe for reading */
  fd_fifo = open(FIFO_HEIKKI, O_RDONLY | O_NDELAY);
  //  if (fd_fifo == NULL) {
  //   printf("Error: could not open fifo.\n"); 
  //  exit(-1);
  // }
#ifdef DEBUG
  printf("now open for reading: %s\n",FIFO_HEIKKI);
  fflush(stdout);
#endif
  
  
  while(1) {

    
    // check named pipe
    // if exists, set heikkis pin high.
    
    if ((num = read(fd_fifo, readbuf, (READBUF_SIZE-1))) > 0) {
      // got a signal 
      readbuf[num] = '\0'; 
#ifdef DEBUG
      printf("Got %i bytes of data: >%s<\n",num, readbuf);
      printf("Set HEIKKI PIN\n");
      fflush(stdout);
#endif
      set_pin(LP_PIN[PIN_HEIKKI]);

      usleep(POLL_TIME);
      clear_pin(LP_PIN[PIN_HEIKKI]);
#ifdef DEBUG
      printf("Clear HEIKKI PIN\n");
      fflush(stdout);
#endif
    }

    
    usleep(POLL_TIME);
    
    // server side
    // fd = open(FIFO_PBAR, O_WRONLY);
    // if ((num = write(fd, s, strlen(s))) == -1)
    // perror("get_serial");
    // close(fd);
    
    //      clear_pin(LP_PIN[PIN_HEIKKI]);
    //#ifdef DEBUG
    //      printf("   Heikki pin cleared. \n");
    //      printf("   cont. \n");
    //      fflush(stdout);
    //#endif   
    
    /*
      if(DEBUG)
      printf("Pin 11: %s\n", pin_is_set(LP_PIN[PIN_TRIGGER]) ? "HIGH":"LOW");
    */
    if (pin_is_set(LP_PIN[PIN_TRIGGER])) {
#ifdef DEBUG
      printf("   Trigger daemon: got a valid trigger. \n");
      fflush(stdout);
#endif

      //      clear_pin(LP_PIN[PIN_HEIKKI]);
      //#ifdef DEBUG
      //      printf("   Heikki pin cleared. \n");
      //      printf("   cont. \n");
      //      fflush(stdout);
      //#endif      

      usleep(TRIGGER_PULSE_WIDTH); /* wait for trigger pulse to end for sure */
      clear_pin(LP_PIN[PIN_RESET]);
      usleep(1000);
      set_pin(LP_PIN[PIN_RESET]);

#ifndef DEBUG
      i = getppid();
      if (i == 1)
	exit(0);
      else
	kill( getppid(), SIGUSR1);
#endif
    }    
  } // end of while(1)
  close(fd_fifo);
  return(0);
}
