/* 
   gcc pbar_main_loop.c logger.c -Wall -lncurses -o pbar_main_loop

   Main data acquisition program.
   
   changelog:

   12.09.2009 No changes, just making this the default version. 
   This code has reentrant fix.

   29.10.2008 this version reads all conversation factors, 
   and enables easy toggling.

   24.10.2008 apadtation to new npbar system.
   - a few fixes, prevent core dump when file doesnt exist

   07.10.2007 complete rewrite of the code.
   - ncurses
   - asynchronous multiprocessing

   TODO: 
   - attach subprocesses   
 
   Niels Bassler <bassler@phys.au.dk>
*/


#include <time.h>
#include <curses.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "logger.h"
#include "integration.h"
#include <fcntl.h>    // O_RDONLY
#include "pbar.h"
#include <sys/stat.h> // S_IFIFO

int current_getch;
int doloop = 1;
static WINDOW *mainwnd;
static WINDOW *screen;
static WINDOW *screen2;
static WINDOW *screen_ccd;
static WINDOW *screen_status;
static WINDOW *screen_acq;
//static WINDOW *screen_hlp;
WINDOW *my_win;

int now_sec, now_min, now_hour, now_day, now_wday, now_month, now_year;
time_t now;
time_t timer0, timer1;
struct tm *now_tm;

#define STAT_LINES 18
#define STAT_COLS 55

char status_buffer[STAT_LINES][STAT_COLS];

char toggle_str[2][10]= {"Disabled","Enabled "};

int toggle=0;

/* Default timings */
#define CURRENT_DEGRADER_MAX 6
#define CURRENT_DEGRADER_MIN 0
int timing_npbar = 50;   // 45 was out of sync October 2006
#define TIMING_NPBAR_MAX 140
#define TIMING_NPBAR_MIN 0
int timing_unidos = 91; 
#define TIMING_UNIDOS_MAX 140
#define TIMING_UNIDOS_MIN 0
int timing_bcm = 4; 
#define TIMING_BCM_MAX 140
#define TIMING_BCM_MIN 0
int timing_sci = 8; 
#define TIMING_SCI_MAX 140
#define TIMING_SCI_MIN 0
int timing_ccd = 94;    
#define TIMING_CCD_MAX 140
#define TIMING_CCD_MIN 0
int timing_ccd_shutter = 4; 
#define TIMING_CCD_SHUTTER_MAX 140
#define TIMING_CCD_SHUTTER_MIN 0

#define DEBUG 0

/* sums */
double npbar_last = 0, npbar_sum = 0;
double unidos_last = 0, unidos_sum = 0;
double bcm_last = 0, bcm_sum = 0;
double sci_last = 0, sci_sum = 0;

/* logics */
int npbar_enabled = TRUE, npbar_arm = FALSE, npbar_done = FALSE;
int unidos_enabled = TRUE, unidos_arm = FALSE, unidos_done = FALSE;
int bcm_enabled = TRUE, bcm_arm = FALSE, bcm_done = FALSE;
int sci_enabled = TRUE, sci_arm = FALSE, sci_done = FALSE;
int ccd_enabled = TRUE, ccd_arm = FALSE, ccd_done = FALSE, ccd_exp = FALSE;
int silent = FALSE;
int first_shot = TRUE;
int pbar_alarm = FALSE;
int pbar_alarm_arm = TRUE;
int pbar_alarm_counter = 0;

/* trigger handeling */
int trigger_lockout = FALSE; /* prevent spurious triggers */
int trigger_lockout_time = TRIGGER_LOCKOUT_TIME;

// NBRE
volatile int got_trigger = FALSE;

/* active device ID for target sum */
int active_target = 0;

long int shot_counter = 0;
char current_time_string[] = "00000000_000000";
float timing_cycle = 0, timing_next = 0, timing_last = 0, timing_current = 0;

/* conversion factors */
float conv_npbar[7], conv_unidos[7], conv_bcm = 0, target_dose[7];
int current_degrader = 0; // 0 for no degrader, 7 for all degraders

#define USAGE_STRING "Usage: %s counter_start_value\n"

void screen_init(void);
static void update_display(void);
void screen_end(void);
void maketime(void);  
static void resize(int sig);
void status_init(void);
void status_write(char *string);
void reload_conversion(void);
//NBRE

static void got_shot_signal(int sig);
void got_shot(void);


int get_npbars(void);
float calc_mean_cycle(void);
void beep_n(const int b);

int main(int argc, char *argv[]) {
  char temp_string[100];
  char temp_filename[100];
  char sci_filename[100]; // remembered for online analysis
  char bcm_filename[100]; 
  char readbuf[READBUF_SIZE];
  FILE *fp;

  //  int status,died;
  int dev_id=0; 
  float dev_msg=0; /* device id and message */

  //  FILE *fp;
  int fd_fifo;
  int num;
  pid_t pid;

  if (argc != 2) {
    printf(USAGE_STRING, argv[0]);
    exit(0);
  }
  printf("\n");


  if (DEBUG)
    printf("DEBUG: setup pbar pipe.\n");

  /* create pipe */
  umask(0);
  if(mknod(FIFO_PBAR, S_IFIFO|0666, 0) != 0) {
    unlink(FIFO_PBAR);
    if(mknod(FIFO_PBAR, S_IFIFO|0666, 0) != 0) {
      printf("Error: could not create fifo.\n"); 
      exit(-1);
    }
    
  }

  if (DEBUG)
    printf("DEBUG: setup pbar pipe...done.\n");



  if (DEBUG)
    printf("DEBUG: open pbar pipe for reading..\n");

  /* open pipe for reading */
  fd_fifo = open(FIFO_PBAR, O_RDONLY | O_NDELAY);

  if (DEBUG)
    printf("DEBUG: done.\n");

  /* start trigger demon */

  if (DEBUG)
    printf("DEBUG: start trigger daemon....\n");

  switch(pid=fork()) {
  case -1: 
    printf("*** ERROR: can't fork\n");
    exit(-1);
    break;
  case 0 : 
    execl(PARPORT_TRIGGER_DAEMON_COMMAND,"",NULL); // this is the code the child runs 
    break;
  default: 
    //    died= wait(&status); // this is the code the parent runs   
    break;
  }  

  if (DEBUG)
    printf("DEBUG: done.\n");


  shot_counter = atoi(argv[1]);
  maketime();


  if (DEBUG)
    printf("DEBUG: init screen..\n");

  status_init();
  screen_init();
  status_write("Acquisition system started.");

  if (DEBUG)
    printf("DEBUG: reload conversion.\n");

  reload_conversion();
  if (DEBUG)
    printf("DEBUG: done.\n");
  

  //  exit(-1);

  status_write("Arming trigger handler.");
  (void) signal(SIGWINCH, resize);    /* resize screen */
  (void) signal(SIGUSR1, got_shot_signal);

  timer0 = time(&now); /* SET TIMER FLAG 0 */
  status_write("Waiting for first trigger signal.");

   while (doloop) {

     if (got_trigger) {
       got_shot();
       got_trigger = FALSE;
     }


     timer1 = time(&now);
     timing_current =  difftime(timer1, timer0);
     if (timing_cycle > 0) {
       timing_next = timing_cycle - timing_current;
     }

     /* re-arm trigger */
     if ((timing_current > trigger_lockout_time) && (trigger_lockout))
       trigger_lockout = FALSE;

     
     /* NPBAR */
     if ((timing_current > timing_npbar) && (npbar_arm) && (npbar_enabled)) {
       status_write("Read NPBARs.");
       //system("./delay2 1&");
       get_npbars();
       npbar_arm = FALSE;
       npbar_done = TRUE;
       //       status_write("NPBAR done.");
     }
     if ((timing_current > timing_unidos) && (unidos_arm) && (unidos_enabled)) {
       status_write("Init UNIDOS read");


       system("./get_serial &");
       // the above causes trouble? 
       // try this instead:
       /*

       switch(pid=fork()) {
       case -1: 
	 status_write("WARNING: can't fork UNIDOS get_serial\n");
	 //exit(-1);
	 break;
       case 0 : 
	 execl("get_serial","",NULL); // this is the code the child runs 
	 exit(0);
	 break;
       default: 
	 //    died= wait(&status); // this is the code the parent runs
	 break;
       }  

       */ // creates a lot of defuct child processes, not good.


       unidos_arm = FALSE;
     }
     if ((timing_current > timing_bcm) && (bcm_arm) && (bcm_enabled)) {
       status_write("Init BCM read");
       sprintf(bcm_filename, "%sBCM_%05li_%s.dat",DATA_PATH_BCM,shot_counter,current_time_string);
       sprintf(temp_string, GET_TRACE_COMMAND, 
	       DEV_BCM, OSC_GPIBADDR, OSC_CHANNEL_BCM, bcm_filename);

       //status_write(temp_string);       
       //printf("\n\nBCM NBNBNB: %s\n",temp_string);
       //exit(1);
       
       system(temp_string);
       bcm_arm = FALSE;
     }
     if ((timing_current > timing_sci) && (sci_arm) && (sci_enabled)) {
       status_write("Init SCI Read");
       sprintf(sci_filename, "%sSCI1_%05li_%s.dat",DATA_PATH_SCI,shot_counter,current_time_string);
       sprintf(temp_string, GET_TRACE_COMMAND, 
	       DEV_SCI, OSC_GPIBADDR, OSC_CHANNEL_SCI, sci_filename);

       //status_write(temp_string);
       //       printf("SCI NBNBNB: %s\n",temp_string);

       system(temp_string);
       sci_arm = FALSE;
     }
     if ((timing_current > timing_ccd) && (ccd_arm) && (ccd_enabled)) {
       status_write("Init CCD Read");
       sprintf(temp_filename, "%sCCD_%05li_%s",DATA_PATH_CCD,shot_counter,current_time_string);
       sprintf(temp_string, CCD_EXPOSE_COMMAND, 
	       temp_filename,timing_ccd_shutter);
       system(temp_string);
       ccd_arm = FALSE;
       ccd_exp = TRUE;
     }
     

     /* check if data is ready from any of the subprocesses */

     if ((num = read(fd_fifo, readbuf, (READBUF_SIZE-1))) > 0) {
       readbuf[num] = '\0'; 
       //sprintf(temp_string,"Got %i bytes of data: >%s<",num, readbuf);
       //status_write(temp_string);

       sscanf(readbuf, "%i %e", &dev_id, &dev_msg);
       switch(dev_id) {
       case DEV_NONE:
	 status_write("No device ready");
	 break;
	 /*       case DEV_NPBAR:
	 status_write("Got NPBARs");
	 get_npbars();
	 npbar_done = TRUE;
	 break;*/
       case DEV_UNIDOS:
	 status_write("Got UNIDOS reply.");
	 if (dev_msg != -1) {
	   unidos_last = dev_msg;
	   unidos_sum += unidos_last;
	 }
	 else
	   status_write("*** UNIDOS proc. returned some error.");


	 fp = fopen(UNIDOS_DATA_FILE, "a");
	 if (fp == NULL) {
	   status_write("*** WARNING: Could not open unidos log data file.");
	   npbar_last = 0;
	 }
	 else {
	 fprintf(fp," %05li %s %e\n",shot_counter,current_time_string,unidos_last);
	 fclose(fp);
	 }
	 unidos_done = TRUE;
	 break;
       case DEV_BCM:
	 status_write("Got BCM reply.");
	 if (bcm_filename != NULL) {
	   bcm_last = bcm_voltage(bcm_filename);
	   if (bcm_last != BCM_ERROR_RETURN_CODE) {
	     if (bcm_last < 0) // fix sign
	       bcm_last *= -1;
	     bcm_sum += bcm_last;
	     bcm_done = TRUE;
	   }
	   else
	     status_write("* ERR: could not open BCM file for read.");
	 }
	 break;
       case DEV_SCI:
	 status_write("Got SCI reply.");
	 if (sci_filename != NULL) {
	   sci_last = sci_integration(sci_filename);
	   if (sci_last == SCI_ERROR_RETURN_CODE) {
	     sci_last = 0;
	     status_write("* ERR: could not open SCI file for read.");
	     sci_done = FALSE;
	   }
	   else {
	     sci_sum += sci_last;
	     sci_done = TRUE;	       
	   }
	 }
	 break;
       case DEV_CCD:
	 status_write("CCD exposure completed.");
	 ccd_exp = FALSE;
	 ccd_done = TRUE;
	 break;
       default:
	 status_write("Warning: invalid return code from FIFO.");
	 break;
	 }
     } // end of subproc check
       

     
     /* Check keys */
     current_getch = getch();	
     switch(current_getch) {
     case (KEY_F(5)+0x20):
       status_write("s+f5");
       break;
     case 'Q':	
       doloop = 0;
       break;
     case (KEY_F0+4):  /* toggle CCD */
       if (ccd_enabled == 0) {
	 ccd_enabled = 1;
	 status_write("CCD acquisition enabled.");
       }
       else {
	 ccd_enabled = 0;
	 status_write("CCD acquisition disabled.");
       }
       break;
     case 'R': /* Reload conversion factors */
       status_write("Reload conversion factors.");
       reload_conversion();	
       break;
     case 'A': /* Pre-arm trigger */
       status_write("Pre-arm trigger.");
       trigger_lockout = FALSE;
       break;
     case 'l': /* Trigger lock */
       ++trigger_lockout_time;
       break;
     case 'L':
       --trigger_lockout_time;
       break;
     case 'k': /* toggle silence */
       silent = !silent;
       break;

     case 'D': /* NPBAR */
       if(current_degrader > CURRENT_DEGRADER_MIN) {
	 current_degrader--;
	 sprintf(temp_string, "Selecting degrader %i.", current_degrader);
	 status_write(temp_string);
       }
       break;
     case 'd':  /* NPBAR */	
       if(current_degrader < CURRENT_DEGRADER_MAX) {
	 current_degrader++;
	 sprintf(temp_string, "Selecting degrader %i.", current_degrader);
	 status_write(temp_string);
       }
       break;


     case 'N': /* NPBAR */
       if(timing_npbar > TIMING_NPBAR_MIN) {
	 timing_npbar--;
	 sprintf(temp_string, "Decreasing NPBAR grace time to %i s.", timing_npbar);
	 status_write(temp_string);
       }
       break;
     case 'n':  /* NPBAR */	
       if(timing_npbar < TIMING_NPBAR_MAX) {
	 timing_npbar++;
	 sprintf(temp_string, "Increasing NPBAR grace time to %i s.", timing_npbar);
	 status_write(temp_string);
       }
       break;
     case 'U': /* UNIDOS */		 
       if(timing_unidos > TIMING_UNIDOS_MIN) {
	 timing_unidos--;
	 sprintf(temp_string, "Decreasing UNIDOS grace time to %i s.", timing_unidos);
	 status_write(temp_string);
       }
       break;
     case 'u': /* UNIDOS */	
       if(timing_unidos < TIMING_UNIDOS_MAX) {
	 timing_unidos++;
	 sprintf(temp_string, "Increasing UNIDOS grace time to %i s.", timing_unidos);
	 status_write(temp_string);
       }
       break;
     case 'B': /* BCM */		 
       if(timing_bcm > TIMING_BCM_MIN) {
	 timing_bcm--;
	 sprintf(temp_string, "Decreasing BCM grace time to %i s.", timing_bcm);
	 status_write(temp_string);
       }
       break;
     case 'b': /* BCM */
       if(timing_bcm < TIMING_BCM_MAX) {
	 timing_bcm++;
	 sprintf(temp_string, "Increasing BCM grace time to %i s.", timing_bcm);
	 status_write(temp_string);
       }
       break;

     case 'S': /* SCI */	
       if(timing_sci > TIMING_SCI_MIN) {
	 timing_sci--;
	 sprintf(temp_string, "Decreasing SCI grace time to %i s.", timing_sci);
	 status_write(temp_string);
       }
       break;

     case 's':  /* SCI */	
       if(timing_sci < TIMING_SCI_MAX) {
	 timing_sci++;
	 sprintf(temp_string, "Increasing SCI grace time to %i s.", timing_sci);
	 status_write(temp_string);
       }
       break;
     case 'E':  /* CCD */	
       if(timing_ccd > TIMING_CCD_MIN) {
	 timing_ccd--;
	 sprintf(temp_string, "Decreasing CCD grace time to %i s.", timing_ccd);
	 status_write(temp_string);
       }
       break;
     case 'e':  /* CCD */	
       if(timing_ccd < TIMING_CCD_MAX) {
	 timing_ccd++;
	 sprintf(temp_string, "Increasing CCD grace time to %i s.", timing_ccd);
	 status_write(temp_string);
       }
       break;
       
     case 'H':  /* CCD */	
       if(timing_ccd_shutter > TIMING_CCD_SHUTTER_MIN) {
	 timing_ccd_shutter--;
	 sprintf(temp_string, "Decreasing CCD exposure time to %i s.", timing_ccd_shutter);
	 status_write(temp_string);
       }
       break;
     case 'h':  /* CCD */	
       if(timing_ccd_shutter < TIMING_CCD_SHUTTER_MAX) {
	 timing_ccd_shutter++;
	 sprintf(temp_string, "Increasing CCD exposure time to %i s.", timing_ccd_shutter);
	 status_write(temp_string);
       }
       break;
     case (KEY_F(5)+12):  /* reset NPBAR sum */
       npbar_sum = 0;
       status_write("Resetting NPBAR sum.");
       break;       
     case (KEY_F(6)+12):  /* reset UNIDOS sum */
       unidos_sum = 0;
       status_write("Resetting UNIDOS sum.");
       break;
     case (KEY_F(7)+12):  /* reset BCM sum */
       bcm_sum = 0;
       status_write("Resetting BCM sum.");
       break;      
     case (KEY_F(8)+12):  /* reset SCI sum */
       sci_sum = 0;
       status_write("Resetting SCI sum.");
       break;  
       
     case (KEY_F(5)):  /* toggle NPBAR */
       if (npbar_enabled == 0) {
	 npbar_enabled = 1;
	 status_write("NPBAR acquisition enabled.");
       }
       else {
	 npbar_enabled = 0;
	 status_write("NPBAR acquisition disabled.");
       }
       break;
       
     case (KEY_F(6)):  /* toggle UNIDOS */
       if (unidos_enabled == 0) {
	 unidos_enabled = 1;
	 status_write("UNIDOS acquisition enabled.");
       }
       else {
	 unidos_enabled = 0;
	 status_write("UNIDOS acquisition disabled.");
       }
       break;
     case (KEY_F(7)):  /* toggle BCM */
       if (bcm_enabled == 0) {
	 bcm_enabled = 1;
	 status_write("BCM acquisition enabled.");
       }
       else {
	 bcm_enabled = 0;
	 status_write("BCM acquisition disabled.");
       }
       break;       
       
     case (KEY_F(8)):  /* toggle SCI */
       if (sci_enabled == 0) {
	 sci_enabled = 1;
	 status_write("SCI acquisition enabled.");
       }
       else {
	 sci_enabled = 0;
	 status_write("SCI acquisition disabled.");
       }
       break;
       
     case 't':  /* toggle Target */
       if (active_target < DEV_BCM) 
	 ++active_target;
       else
	 active_target = DEV_NONE;
       break;       

     case ' ':
       if (pbar_alarm_arm) {
	 pbar_alarm_arm = FALSE;
	 status_write("Trigger alarm silenced.");
       }
       else {
	 pbar_alarm_arm = TRUE;
	 status_write("Trigger alarm armed.");
       }
       break;
     default:
       //       if (current_getch > 0) {
       //       sprintf(temp_string,"%i",current_getch);
       //       status_write(temp_string);}
       break;

     } // end of key handler

     maketime();
     update_display();

     if ((timing_current > (timing_cycle+4)) && (timing_cycle > 0)) 
       pbar_alarm = TRUE;
     else 
       pbar_alarm = FALSE;     

     if (pbar_alarm && pbar_alarm_arm) {
       pbar_alarm_counter++;
       if (pbar_alarm_counter >= 13) {
	 pbar_alarm_counter = 0;
	 beep();
       }
     }
     usleep(50000);
   }
   screen_end();
   close(fd_fifo);
   unlink(FIFO_PBAR);
   printf("TEST ENDS\n");
   return 0;
}
	
void screen_init(void) {
   mainwnd = initscr();
   keypad(stdscr,TRUE);
   if(has_colors() == FALSE)     {	
     endwin();
     printf("Your terminal does not support color\n");
     exit(1);
   }

   start_color();
   clear();
   
   init_pair(COLOR_BLACK, COLOR_BLACK, COLOR_BLACK);
   init_pair(COLOR_GREEN, COLOR_GREEN, COLOR_BLACK);
   init_pair(COLOR_RED, COLOR_RED, COLOR_BLACK);
   init_pair(COLOR_CYAN, COLOR_CYAN, COLOR_BLACK);
   init_pair(COLOR_WHITE, COLOR_WHITE, COLOR_BLACK);
   init_pair(COLOR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
   init_pair(COLOR_BLUE, COLOR_BLUE, COLOR_BLACK);
   init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLACK);   

   noecho();
   cbreak();
   nodelay(mainwnd, TRUE);
   refresh(); // 1)
   wrefresh(mainwnd);

   screen = newwin(14, 28, 1, 1);
   screen2 = newwin(7, 27, 1, 30);
   screen_ccd = newwin(7, 27, 8, 30);
   screen_status = newwin(STAT_LINES+1, STAT_COLS+1, 15, 1);
   screen_acq = newwin(33, 86,  1, 58);
   //   screen_hlp = newwin(10, 143, 34, 1);

   box(screen, ACS_VLINE, ACS_HLINE);
   box(screen2, ACS_VLINE, ACS_HLINE);
   box(screen_ccd, ACS_VLINE, ACS_HLINE);
   box(screen_status, ACS_VLINE, ACS_HLINE);
   box(screen_acq, ACS_VLINE, ACS_HLINE);
   //   box(screen_hlp, ACS_VLINE, ACS_HLINE);
}

static void update_display(void) {
  int i=0;
  // postions 
  static const int pn=1, pu=22, pb=43, ps=64;
  static float delta_ts=0;
  
  curs_set(0);
  wattron(screen, A_BOLD);
  mvwprintw(screen,1,1,"---- MAIN TIMER ---------");
  wattroff(screen, A_BOLD);
  mvwprintw(screen,3,4,"TIME: %02d:%02d:%02d", now_hour, now_min, now_sec);
  mvwprintw(screen,5,4,"DATE: %d-%d-%d", now_day, now_month, now_year);
  mvwprintw(screen,7,4,"LAST SHOT#: %7i",shot_counter);
  mvwprintw(screen,9,4,"ID: %s", current_time_string);
  mvwprintw(screen,11,4,"PID: %i", (int) getpid());

  wrefresh(screen);
  
  wattron(screen2, A_BOLD);
  mvwprintw(screen2,1,1,"---- SPILL TIMING ------");
  wattroff(screen2, A_BOLD);
  mvwprintw(screen2,3,4,"Last:   %6.0f s",timing_current);
  mvwprintw(screen2,4,4,"Next:   %6.0f s",timing_next);
  mvwprintw(screen2,5,4,"Cycle : %6.1f s",timing_cycle);
  wrefresh(screen2);

  /* CCD window -----------------------------------------------------------------------*/
  wattron(screen_ccd, A_BOLD);
  mvwprintw(screen_ccd,1,1,"---- CCD  --------------");
  wattroff(screen_ccd,A_BOLD);  
  mvwprintw(screen_ccd,3,4, "Status:");
  wattron(screen_ccd, A_BOLD);
  if (ccd_enabled) 
    wattron(screen_ccd,COLOR_PAIR(COLOR_GREEN));
  else 
    wattron(screen_ccd,COLOR_PAIR(COLOR_RED));
  
  mvwprintw(screen_ccd,3,4+9, " %s", toggle_str[ccd_enabled]);
  if (ccd_enabled) 
    wattroff(screen_ccd,COLOR_PAIR(COLOR_GREEN));
  else
    wattroff(screen_ccd,COLOR_PAIR(COLOR_RED));
  wattroff(screen_ccd, A_BOLD);
  

  if (ccd_exp) {
    wattron(screen_ccd,COLOR_PAIR(COLOR_MAGENTA));
    wattron(screen_ccd, A_BOLD);
    mvwprintw(screen_ccd,3,4+9, " Exposing");
    wattroff(screen_ccd,A_BOLD);
    wattroff(screen_ccd,COLOR_PAIR(COLOR_MAGENTA));
  }
  
  
  //   mvwprintw(screen_ccd,3,4,"Status: %s", toggle_str[ccd_enabled]);
  mvwprintw(screen_ccd,4,4,"Exp.  : %3i s [+/-]e",timing_ccd);
  mvwprintw(screen_ccd,5,4,"Sht.  : %3i s [+/-]h",timing_ccd_shutter);
  wrefresh(screen_ccd);
  
  wattron(screen_status, A_BOLD);
  mvwprintw(screen_status,1,1,"----- STATUS ---------------------------------------- ");
  wattroff(screen_status,A_BOLD);
  for (i=0;i<(STAT_LINES-2);i++)
    mvwprintw(screen_status,(i+2),1,status_buffer[i]);
  wrefresh(screen_status);
  
  wattron(screen_acq, A_BOLD);
  mvwprintw(screen_acq,1,pn,"---- NPBAR -------");
  mvwprintw(screen_acq,1,pu,"---- UNIDOS ------");
  mvwprintw(screen_acq,1,pb,"---- BCM ---------");
  mvwprintw(screen_acq,1,ps,"---- SCI ---------");
  wattroff(screen_acq,A_BOLD);
  
  /* NPBAR ------------------------------------------------------------------------- */
  mvwprintw(screen_acq,3,pn, " Status :");
  wattron(screen_acq, A_BOLD);
  if (npbar_enabled) 
    wattron(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else 
    wattron(screen_acq,COLOR_PAIR(COLOR_RED));
  
  mvwprintw(screen_acq,3,pn+9, " %s", toggle_str[npbar_enabled]);
  if (npbar_enabled) 
    wattroff(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else
    wattroff(screen_acq,COLOR_PAIR(COLOR_RED));
  wattroff(screen_acq, A_BOLD);
  
  
  mvwprintw(screen_acq,5,pn, " Last   :");
  mvwprintw(screen_acq,6,pn, " Sum    :");
  mvwprintw(screen_acq,7,pn, " Target : %7.2f",target_dose[current_degrader]/conv_npbar[current_degrader]);
  mvwprintw(screen_acq,9,pn, "    x 1e7 pbar ");
  mvwprintw(screen_acq,11,pn, " Acq. grace time:");
  mvwprintw(screen_acq,12,pn,"  %3i s [+/-]n",timing_npbar);
  
  mvwprintw(screen_acq,14,pn," Plateau Dose (Gy):");
  mvwprintw(screen_acq,16,pn," Sum   :");
  
  mvwprintw(screen_acq,18,pn," Conv. factor:");
  mvwprintw(screen_acq,19,pn,"  %7.3e",conv_npbar[current_degrader]);
  mvwprintw(screen_acq,20,pn,"  Gy/1e7pbar");
  
  if (npbar_done)
    wattron(screen_acq, A_BOLD);
  mvwprintw(screen_acq,5,pn+9, " %7.2f",npbar_last);   
  mvwprintw(screen_acq,6,pn+9, " %7.2f",npbar_sum);
  mvwprintw(screen_acq,16,pn+9, " %7.3f",npbar_sum * conv_npbar[current_degrader]);
  if (npbar_done)
    wattroff(screen_acq, A_BOLD);
  
  
  /* UNIDOS -------------------------------------------------------------------*/  
  mvwprintw(screen_acq,3,pu, " Status :");
  wattron(screen_acq, A_BOLD);
  if (unidos_enabled) 
    wattron(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else 
    wattron(screen_acq,COLOR_PAIR(COLOR_RED));
  
  mvwprintw(screen_acq,3,pu+9, " %s", toggle_str[unidos_enabled]);
  if (unidos_enabled) 
    wattroff(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else
    wattroff(screen_acq,COLOR_PAIR(COLOR_RED));
  wattroff(screen_acq, A_BOLD);
  
  // mvwprintw(screen_acq,3,pu, " Status : %s", toggle_str[unidos_enabled]);
  mvwprintw(screen_acq,5,pu, " Last   :");
  mvwprintw(screen_acq,6,pu, " Sum    :");
  mvwprintw(screen_acq,7,pu, " Target : %7.2e",target_dose[current_degrader]/conv_unidos[current_degrader]);
  mvwprintw(screen_acq,9,pu, "      Coulomb ");
  mvwprintw(screen_acq,12,pu,"  %3i s [+/-]u",timing_unidos);
  
  mvwprintw(screen_acq,14,pu," Last  :");
  mvwprintw(screen_acq,16,pu," Sum   :");
  
  mvwprintw(screen_acq,19,pu,"  %7.3e",conv_unidos[current_degrader]);
  mvwprintw(screen_acq,20,pu,"  Gy/C");
  
  if (unidos_done)
    wattron(screen_acq, A_BOLD);
  mvwprintw(screen_acq,5,pu+9,  " %8.2e",unidos_last);   
  mvwprintw(screen_acq,6,pu+9,  " %8.2e",unidos_sum);
  mvwprintw(screen_acq,14,pu+9, "%7.3f",unidos_last * conv_unidos[current_degrader]);
  mvwprintw(screen_acq,16,pu+9, "%7.3f",unidos_sum * conv_unidos[current_degrader]);
  if (unidos_done)
    wattroff(screen_acq, A_BOLD);
  
  
  /* BCM -------------------------------------------------------------------*/
  mvwprintw(screen_acq,3,pb, " Status :");
  wattron(screen_acq, A_BOLD);
  if (bcm_enabled) 
    wattron(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else 
    wattron(screen_acq,COLOR_PAIR(COLOR_RED));
  
  mvwprintw(screen_acq,3,pb+9, " %s", toggle_str[bcm_enabled]);
  if (bcm_enabled) 
    wattroff(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else
    wattroff(screen_acq,COLOR_PAIR(COLOR_RED));
  wattroff(screen_acq, A_BOLD);
  //mvwprintw(screen_acq,3,pb, " Status : %s", toggle_str[bcm_enabled]);
  mvwprintw(screen_acq,5,pb, " Last   :");
  mvwprintw(screen_acq,6,pb, " Sum    :");
  mvwprintw(screen_acq,7,pb, " Target : %7.2e",target_dose[current_degrader]/conv_bcm);
  mvwprintw(screen_acq,9,pb, "       Volt    ");
  mvwprintw(screen_acq,12,pb,"  %3i s [+/-]b",timing_bcm);
  
  mvwprintw(screen_acq,16,pb," Sum   :");
  
  mvwprintw(screen_acq,19,pb,"  %7.3e",conv_bcm);
  mvwprintw(screen_acq,20,pb,"  Gy/V");
  
  if (bcm_done)
    wattron(screen_acq, A_BOLD);
  mvwprintw(screen_acq,5,pb+9,  " %8.2f",bcm_last);   
  mvwprintw(screen_acq,6,pb+9,  " %8.2f",bcm_sum);
  mvwprintw(screen_acq,16,pb+9, "%7.3f",bcm_sum * conv_bcm);
  if (bcm_done)
    wattroff(screen_acq, A_BOLD);
  
  /* SCI -------------------------------------------------------------------*/
  mvwprintw(screen_acq,3,ps, " Status :");
  wattron(screen_acq, A_BOLD);
  if (sci_enabled) 
    wattron(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else 
    wattron(screen_acq,COLOR_PAIR(COLOR_RED));
  
  mvwprintw(screen_acq,3,ps+9, " %s", toggle_str[sci_enabled]);
  if (sci_enabled) 
    wattroff(screen_acq,COLOR_PAIR(COLOR_GREEN));
  else
    wattroff(screen_acq,COLOR_PAIR(COLOR_RED));
  wattroff(screen_acq, A_BOLD);
  
  //mvwprintw(screen_acq,3,ps, " Status : %s", toggle_str[sci_enabled]);
  mvwprintw(screen_acq,5,ps, " Last  :");
  mvwprintw(screen_acq,6,ps, " Sum   :");
  mvwprintw(screen_acq,9,ps, "     Voltsec   ");
  mvwprintw(screen_acq,12,ps,"   %3i s [+/-]s",timing_sci);
  
  if (sci_done)
    wattron(screen_acq, A_BOLD);
  mvwprintw(screen_acq,5,ps+9," %8.2e ",sci_last);   
  mvwprintw(screen_acq,6,ps+9," %8.2e ",sci_sum);
  if (sci_done)
    wattroff(screen_acq, A_BOLD);
  
  
  
  /* link the target dose to a device */
  switch(active_target) {
  case DEV_NPBAR:
    wattron(screen_acq, A_BOLD);
    mvwprintw(screen_acq,17,pn," Target: %7.3f Gy",target_dose[current_degrader]);
    wattroff(screen_acq, A_BOLD);
    mvwprintw(screen_acq,17,pu,"                   ");
    mvwprintw(screen_acq,17,pb,"                   ");
    delta_ts = ((target_dose[current_degrader]/conv_npbar[current_degrader]) - npbar_sum)/ (npbar_last / (float) timing_cycle);
    break;
  case DEV_UNIDOS:
    mvwprintw(screen_acq,17,pn,"                   ");
    wattron(screen_acq, A_BOLD);
    mvwprintw(screen_acq,17,pu," Target: %7.3f Gy",target_dose[current_degrader]);
    wattroff(screen_acq, A_BOLD);
    mvwprintw(screen_acq,17,pb,"                   ");
    delta_ts = ((target_dose[current_degrader]/conv_unidos[current_degrader]) - unidos_sum)/ (unidos_last / (float) timing_cycle);
    break;
  case DEV_BCM:
    mvwprintw(screen_acq,17,pn,"                   ");
    mvwprintw(screen_acq,17,pu,"                   ");
    wattron(screen_acq, A_BOLD);
    mvwprintw(screen_acq,17,pb," Target: %7.3f Gy",target_dose[current_degrader]);
    wattroff(screen_acq, A_BOLD);
    delta_ts = ((target_dose[current_degrader]/conv_bcm) - bcm_sum)/ (bcm_last / (float) timing_cycle);
    break;
  default:
    mvwprintw(screen_acq,17,pn,"                   ");
    mvwprintw(screen_acq,17,pu,"                   ");
    mvwprintw(screen_acq,17,pb,"                   ");
    delta_ts = 0;
    break;
  }
  
  /* ETA */
  if (active_target != 0) {
    wattron(screen_acq, A_BOLD);
    if ((delta_ts/3600.0) > 1)
      mvwprintw(screen_acq,31,55, "Est. time left: %4.2f hours.  ",delta_ts/3600.0);
    else if (delta_ts < 0)
      if (delta_ts > -3600)
	mvwprintw(screen_acq,31,55, "Overdue: %4.0f minutes.       ",delta_ts/60.0);
      else
	mvwprintw(screen_acq,31,55, "                            ");
    else
      mvwprintw(screen_acq,31,55, "Est. time left: %4.0f minutes.",delta_ts/60.0);
    wattroff(screen_acq, A_BOLD);
  }
  else
    mvwprintw(screen_acq,31,55, "                            ");
  

  mvwprintw(screen_acq,25,1, "    Current degrader selected: %i [+/-]d", current_degrader);



  if (pbar_alarm) {
    wattron(screen_acq, A_BOLD);
    mvwprintw(screen_acq,27,1, "    Warning: Trigger delayed!");
    wattroff(screen_acq, A_BOLD);
  }
  else
    mvwprintw(screen_acq,27,1, "                             ");
  
  
  mvwprintw(screen_acq,28,1, "    Trigger lockout time: %i s [+/-]l", trigger_lockout_time);
  if (!trigger_lockout) {
    mvwprintw(screen_acq,29,1, "    Trigger armed");
  } else
    mvwprintw(screen_acq,29,1, "                 ");

  if (silent) {
    mvwprintw(screen_acq,30,1, "    Silent mode");
  }
  else
    mvwprintw(screen_acq,30,1, "               ");
  wrefresh(screen_acq);
  //  wrefresh(screen_hlp);
  refresh();
}

void screen_end(void) {
   endwin();
}

void maketime(void) {
	// Get the current date/time
	now = time (NULL);
	now_tm = localtime (&now);
	now_sec = now_tm->tm_sec;
	now_min = now_tm->tm_min;
	now_hour = now_tm->tm_hour;
	now_day = now_tm->tm_mday;
	now_wday = now_tm->tm_wday;
	now_month = now_tm->tm_mon + 1;
	now_year = now_tm->tm_year + 1900;
}

static void resize(int sig)
{
  endwin();
  screen_init();
  update_display();
}

void status_init(void) {
  int i=0;
  for (i = 0; i < STAT_LINES; i++) 
    sprintf(status_buffer[i]," ");
}

void status_write(char *string) {
  int i=0;
  static int current_line = 0;
  char string_out[STAT_COLS];
  char temp_string[100];

  maketime();
  sprintf(string_out,"%02d:%02d:%02d ", now_hour, now_min, now_sec);  
  
  time(&now);
  strftime (temp_string, 22,"%Y.%m.%d %H:%M:%S", localtime (&now));

  logger(shot_counter, temp_string, 0, string);


  if (strlen(string) > (STAT_COLS-10))
    strcpy(temp_string, "(WARNING: truncated string)");
  else 
    sprintf(temp_string,"%s%-99s", string_out, string);
  
  strncpy(string_out, temp_string, STAT_COLS-1); /* truncate to proper length, avoid core */
  // strcat(string_out,string);

  if (current_line == (STAT_LINES-2)) {
    for (i = 0; i < (STAT_LINES-3); i++)
      strncpy(status_buffer[i], status_buffer[i+1],STAT_COLS-1);
    strncpy(status_buffer[STAT_LINES-3],string_out,STAT_COLS-1);
  }
  else {
    strncpy(status_buffer[current_line], string_out,STAT_COLS-1);
    current_line++;
  }
  
}
 
void reload_conversion(void){
  FILE *fp;
  char line[200];

  fp = fopen(PBAR_CONV_DEGR_FILENAME,"r");
  if (fp == NULL) {
    status_write("Error: Could not open conv. file.");
  }
  else {
    if (fgets(line, 200, fp) != NULL) {
      sscanf(line,"%g %g %g %g %g %g %g", &conv_npbar[0],&conv_npbar[1],&conv_npbar[2],&conv_npbar[3],&conv_npbar[4],&conv_npbar[5],&conv_npbar[6]);
      status_write("Updating NPBAR conversion factor.");
    }
    else
      status_write("Error: conv. file, invalid format.");
    
    if (fgets(line, 200, fp) != NULL) {
      sscanf(line, "%g %g %g %g %g %g %g", &conv_unidos[0],&conv_unidos[1],&conv_unidos[2],&conv_unidos[3],&conv_unidos[4],&conv_unidos[5],&conv_unidos[6]);
      status_write("Updating UNIDOS conversion factor.");}
    else
      status_write("Error: conv. file, invalid format.");


    if (fgets(line, 200, fp) != NULL) {
      sscanf(line, "%g", &conv_bcm);
      status_write("Updating BCM conversion factor.");}
    else
      status_write("Error: conv. file, invalid format.");

    if (fgets(line, 200, fp) != NULL) {
      sscanf(line, "%g %g %g %g %g %g %g", &target_dose[0],&target_dose[1],&target_dose[2],&target_dose[3],&target_dose[4],&target_dose[5],&target_dose[6]);
      status_write("Updating requested plateau dose.");}
    else
      status_write("Error: conv. file, invalid format.");
  }
  fclose(fp);
}


// NBRE
void got_shot_signal(int sig) {
  got_trigger = TRUE;
}

void got_shot(void) {
  char temp_string[100];
  
  if (trigger_lockout) {
    sprintf(temp_string,"Warning: early tigger t = %.0f sec.",timing_current);
    status_write(temp_string);
    beep_n(2);
    return;
  }
  /* protect against spurious trigger */
  trigger_lockout = TRUE;


  switch(active_target) {
  case DEV_NPBAR:
    if((npbar_sum * conv_npbar[current_degrader]) >= target_dose[current_degrader]) 
      beep_n(5);
    break;
  case DEV_UNIDOS:
    if((unidos_sum * conv_unidos[current_degrader]) >= target_dose[current_degrader])
      beep_n(5);
    break;
  case DEV_BCM:
    if((bcm_sum * conv_bcm) >= target_dose[current_degrader])
      beep_n(5);
    break;
  default:
    break;
  }
  beep_n(1);
  
  /* record elapsed time*/
  timing_last = timing_current;
  /* only after first shot */
  if (!first_shot)
    timing_cycle = calc_mean_cycle();
  first_shot = FALSE;

  /* save all aquired data */
  data_logger(shot_counter, current_time_string,
	      npbar_last, unidos_last, bcm_last, sci_last);
  /* increase SHOT_COUNTER by one */
  ++shot_counter;

  sprintf(temp_string,"*** Got spill number %.0f ! dt=%.1f", (float) shot_counter, (float) timing_last); 
  status_write(temp_string);
  /* What time is it ? Generate time stamp string from this. */
  timer0 = time(&now); /* SET TIMER FLAG 0 */
  time(&now);
  strftime (current_time_string, 17,"%Y%m%d_%H%M%S", localtime (&now));
  // we set the timestamp.
  // now return to main loop, and keep refreshing UI while checking grace times.
  // if grace times are ok, then init the acq process
  // and keep polling until data are recieved.
  if(npbar_enabled)
    npbar_arm = TRUE;
  else
    npbar_arm = FALSE;
  if(unidos_enabled)
    unidos_arm = TRUE;
  else
    unidos_arm = FALSE;
  if(bcm_enabled)
    bcm_arm = TRUE;
  else
    bcm_arm = FALSE;
  if(sci_enabled)
    sci_arm = TRUE;
  else
    sci_arm = FALSE;
  /* ccd should be armed anyway */
  ccd_arm = TRUE;
  
  npbar_done = FALSE;
  unidos_done = FALSE;
  bcm_done = FALSE;
  sci_done = FALSE;



}



int get_npbars(void) {

  FILE *fp;

  char string[0xFF],buf[0xFF], npbar_time_string[0xFF];
  double mytime;
  //float a,b,c,d,e;
  char *endp;
  time_t tm;
  float npbar = 0;

  fp = fopen(NPBAR_FILENAME_LAST_SHOT, "r");

  if (fp == NULL) {
    status_write("*** WARNING: Could not open npbar source data file.\n");
    npbar_last = 0;
  }
  else {
    // go to the end of file
    while( fgets(string,0xFF,fp) ) {
    }
    //    sscanf(string,"%s %f %f %f %f %f %f",buf,&a,&b,&c,&d,&npbar,&e);
    sscanf(string,"%s %*f %*f %*f %*f %f %*f",buf,&npbar);
    mytime = strtod(buf,&endp);
    mytime /= 1000.0; // convert to seconds
    tm = (time_t) mytime;

    // convert unix timestamp from npbar to npbar time stamp
    strftime (npbar_time_string, 17,"%Y%m%d_%H%M%S", localtime(&tm));        
    //printf("%s %f - %f\n",buf2, mytime,npbar);
    //fp2 = fopen(NPBAR_FILENAME_LAST_SHOT,"w");
    //      fprintf("");
  }
  fclose(fp);

  fp = fopen(NPBAR_DATA_FILE, "a");
  if (fp == NULL) {
    status_write("*** WARNING: Could not open npbar log data file.\n");
    npbar_last = 0;
  }
  else{
  fprintf(fp," %05li %s %s %.2f\n",shot_counter,current_time_string,npbar_time_string,npbar);
  fclose(fp);
  }

  if(npbar < NPBAR_NOISE_THRESHOLD)
    npbar = 0;

  npbar_last = npbar;
  npbar_sum += npbar;

  return(0);
}


float calc_mean_cycle(void) {
  float mean;
  static float c[CYCLES_AVR];
  static int init = 1;
  static int index = 0;

  int i,j=0;
  float c_sum = 0;

  /* empty cycle placeholders */
  if (init == 1) {
    for (i = 0; i<CYCLES_AVR; i++) 
      c[i] = 0;
    init = 0;
  }

  // add  last value (overwrite old ones)
  c[index] = timing_last;
  ++index;
  if (index >= CYCLES_AVR)
    index=0;

  
  // calculate average
  for (i = 0; i<CYCLES_AVR; i++) {    
    if (c[i] !=0) {
      c_sum += c[i];
      j++;
    }
  }
  return mean = c_sum / (float) j;

}

void beep_n(const int b) {
  int i;
  if (!silent)
    {
      for (i=0; i<b; i++) {
	beep();
	usleep(100000);
      }
    }
}
